function [valid]=verifySimplex(bx0,bx1,bx2,PHI_0,PHI_1,PHI_2,alphaM,epsilon)
%   INPUT DATA
% bx0, bx1, bx2: coordinates of the nodes that identify the frame
% PHI_0, PHI_1, PHI_2: orthonormal frames attached to the nodes above
%  (respectively)
% alphaM, epsilon: output data of int_estimates_cod.m
%
%   OUTPUT DATA
% valid: logical variable, true if verification was succesfull, false
%  otherwise
%
%   NOTES
% We choose the node bx0 to be the node that minimizes the variation
% between the frame attached to it and the frames attached to the two
% other nodes (see below, lines 28-43)
%
% The attempted proof is based on bx0, that is the Newton-like operator
% is of the form T_s(x) = x - A*F_s(x), where A is an approximate inverse
% of the Jacobian linear operator DF_s(bx0).

bx0 = intval(bx0);
bx1 = intval(bx1);
bx2 = intval(bx2);
PHI_0 = intval(PHI_0);
PHI_1 = intval(PHI_1);
PHI_2 = intval(PHI_2);

%% reorder the nodes so that the node on which we base the verificarion 
%  minimizes the distance between reference frames at the endpoints of the
%  edges attached to it
d01=mag(norm(PHI_0-PHI_1,'fro'));
d02=mag(norm(PHI_0-PHI_2,'fro'));
d12=mag(norm(PHI_1-PHI_2,'fro'));
[~,dMaxInd]=max([d12 d02 d01]);
if dMaxInd>1
    bx_mat=[bx0 bx1 bx2];
    bx_mat(:,[1 dMaxInd])=bx_mat(:,[dMaxInd 1]);
    PHI_mat=[PHI_0 PHI_1 PHI_2];
    PHI_mat(:,[1 2 2*dMaxInd-1 2*dMaxInd])=PHI_mat(:,[2*dMaxInd-1 2*dMaxInd 1 2]);
    bx0=bx_mat(:,1); bx1=bx_mat(:,2); bx2=bx_mat(:,3);
    PHI_0=PHI_mat(:,[1 2]); PHI_1=PHI_mat(:,[3 4]); PHI_2=PHI_mat(:,[5 6]);
end
%%

lambda0=bx0(1);
lambda1=bx1(1);
lambda2=bx2(1);

u0=bx0(3:end);
u1=bx1(3:end);
u2=bx2(3:end);

m=length(u0);
M=3*m-2;
s=2;
valid=false;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Beginning of the part involving interval arithmetic %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

i0=intval('0');
i1=intval('1');
i2=intval('2');
i3=intval('3');
i6=intval('6');
i9=intval('9');
iL=midrad(3.14159265358979323,1e-16);
iM=intval(M);
is=intval(s);

Lk2=(intval(0:M-1)'*iL).^2;
u0M=intval([u0;zeros(M-m,1)]);
u1M=intval([u1;zeros(M-m,1)]);
u2M=intval([u2;zeros(M-m,1)]);
tuM=intval([1 (1:M-1)]').^(-intval(s));

lambda0=intval(lambda0);
lambda1=intval(lambda1);
lambda2=intval(lambda2);

Am=DF0_ch(mid(bx0),mid(PHI_0(:,1)),mid(PHI_0(:,2)))\eye(m+2);
iAm=intval(Am);

int_Df0M=int_Df_ch(intval([bx0;zeros(M-m,1)]));

int_DF0M=intval(zeros(m+2));
int_DF0M(1,:)=intval(PHI_0(:,1)');
int_DF0M(2,:)=intval(PHI_0(:,2)');
int_DF0M(3:end,:)=int_Df0M(1:m,1:m+2);

%%%%%%%%%%%
%%% Y_F %%%
%%%%%%%%%%%

xi=lambda0+infsup(0,1)*(lambda1-lambda0)+infsup(0,1)*(lambda2-lambda0);

y00=abs(int_f_ch(intval([bx0;zeros(M-m,1)])));
y10=abs(int_Df0M*(intval([bx1-bx0;zeros(M-m,1)])));
y01=abs(int_Df0M*(intval([bx2-bx0;zeros(M-m,1)])));
y20=abs(Lk2/(lambda0^2)*(lambda1-lambda0).*(u1M-u0M)-Lk2/(xi^3)*(lambda1-lambda0)^2.*u0M-i3*int_cubicFFT(u0M,u1M-u0M,u1M-u0M));
y11=abs(Lk2/(lambda0^2)*(lambda2-lambda0).*(u1M-u0M)+Lk2/(lambda0^2)*(lambda1-lambda0).*(u2M-u0M) ...
    -i2*Lk2/(xi^3)*(lambda2-lambda0)*(lambda1-lambda0).*u0M-i6*int_cubicFFT(u0M,u2M-u0M,u1M-u0M));
y02=abs(Lk2/(lambda0^2)*(lambda2-lambda0).*(u2M-u0M)-Lk2/(xi^3)*(lambda2-lambda0)^2.*u0M-i3*int_cubicFFT(u0M,u2M-u0M,u2M-u0M));
y30=abs(-Lk2/(xi^3)*(lambda1-lambda0)^2.*(u1M-u0M)-int_cubicFFT(u1M-u0M,u1M-u0M,u1M-u0M));
y21=abs(-i2*Lk2/(xi^3)*(lambda2-lambda0)*(lambda1-lambda0).*(u1M-u0M)-Lk2/(xi^3)*(lambda1-lambda0)^2.*(u2M-u0M) ...
    -i3*int_cubicFFT(u2M-u0M,u1M-u0M,u1M-u0M));
y12=abs(-Lk2/(xi^3)*(lambda2-lambda0)^2.*(u1M-u0M)-i2*Lk2/(xi^3)*(lambda2-lambda0)*(lambda1-lambda0).*(u2M-u0M) ...
    -i3*int_cubicFFT(u2M-u0M,u2M-u0M,u1M-u0M));
y03=abs(-Lk2/(xi^3)*(lambda2-lambda0)^2.*(u2M-u0M)-int_cubicFFT(u2M-u0M,u2M-u0M,u2M-u0M));

y = y00+y10+y01+y20+y11+y02+y30+y21+y12+y03;

Y_F=[intval(abs(Am))*[i0;i0;y(1:m)];y(m+1:end)./abs(i1-(i1/lambda0)*Lk2(m+1:end))];

%%%%%%%%%%%
%%% Z_F %%%
%%%%%%%%%%%

A0=intval(max(mag(intval(abs(u0)).*(intval([1 (1:m-1)]).^intval(s))')));
A1_0=intval(max(mag(intval(abs(u1-u0)).*(intval([1 (1:m-1)]).^intval(s))')));
A2_0=intval(max(mag(intval(abs(u2-u0)).*(intval([1 (1:m-1)]).^intval(s))')));

% [alphaM,epsilon]=int_estimates_cod(s,M);

z101=abs(Lk2.*(u2M-u0M)./lambda0^2-i2*Lk2*(lambda2-lambda0).*u0M/lambda0^3) ... 
     +Lk2*abs(lambda2-lambda0).*tuM/lambda0^2+i6*(int_cubicFFT(abs(u0M),abs(u2M-u0M),tuM)+i3*A0*A2_0*epsilon);
 
z102=abs(-i2*Lk2*(lambda2-lambda0).*(u2M-u0M)/lambda0^3+i3*Lk2*(lambda2-lambda0)^2.*u0M/xi^4) ... 
     +abs(Lk2*abs(lambda2-lambda0)^2.*tuM/xi^3)+i3*(int_cubicFFT(abs(u2M-u0M),abs(u2M-u0M),tuM)+i3*A2_0*A2_0*epsilon);

z103=abs(i3*Lk2*(lambda2-lambda0)^2.*(u2M-u0M)/xi^4);

z110=abs(Lk2.*(u1M-u0M)/lambda0^2-i2*Lk2*(lambda1-lambda0).*u0M/lambda0^3) ... 
     +Lk2*abs(lambda1-lambda0).*tuM/lambda0^2+i6*(int_cubicFFT(abs(u0M),abs(u1M-u0M),tuM)+i3*A0*A1_0*epsilon);
 
z111=intval(mag(-i2*Lk2*(lambda2-lambda0).*(u1M-u0M)/lambda0^3-i2*Lk2*(lambda1-lambda0).*(u2M-u0M)/lambda0^3 ...
         +i6*Lk2*(lambda2-lambda0)*(lambda1-lambda0).*u0M/xi^4)) ...
     +intval(mag(i2*Lk2*abs(lambda2-lambda0)*abs(lambda1-lambda0).*tuM/xi^3)) ... 
     +i6*(int_cubicFFT(abs(u2M-u0M),abs(u1M-u0M),tuM)+i3*A2_0*A1_0*epsilon);

z112=intval(mag(i3*Lk2*(lambda2-lambda0)^2.*(u1M-u0M)/xi^4+i6*Lk2*(lambda2-lambda0)*(lambda1-lambda0).*(u2M-u0M)/xi^4));

z120=intval(mag(-i2*Lk2*(lambda1-lambda0).*(u1M-u0M)/lambda0^3+i3*Lk2*(lambda1-lambda0)^2.*u0M/xi^4)) ...
     +intval(mag(Lk2*(lambda1-lambda0)^2.*tuM/xi^3))+i3*(int_cubicFFT(abs(u1M-u0M),abs(u1M-u0M),tuM)+i3*A1_0*A1_0*epsilon);
 
z121=intval(mag(i6*Lk2*(lambda2-lambda0)*(lambda1-lambda0).*(u1M-u0M)/xi^4+i3*Lk2*(lambda1-lambda0)^2.*(u2M-u0M)/xi^4));

z130=intval(mag(i3*Lk2*(lambda1-lambda0)^2.*(u1M-u0M)/xi^4));

z1=z101+z102+z103+z110+z111+z112+z120+z121+z130;

z200=abs(i2*Lk2.*u0M/lambda0^3)+i6*(int_cubicFFT(abs(u0M),tuM,tuM)+i3*A0*epsilon);

z201=intval(mag(-i2*Lk2.*(u2M-u0M)/lambda0^3+i6*Lk2*(lambda2-lambda0).*u0M/xi^4)) ...
     +i6*(int_cubicFFT(abs(u2M-u0M),tuM,tuM)+i3*A2_0*epsilon);
 
z202=intval(mag(i6*Lk2*(lambda2-lambda0).*(u2M-u0M)/xi^4));

z210=intval(mag(-i2*Lk2.*(u1M-u0M)/lambda0^3+i6*Lk2*(lambda1-lambda0).*u0M/xi^4)) ...
     +i6*(int_cubicFFT(abs(u1M-u0M),tuM,tuM)+i3*A1_0*epsilon);
 
z211=intval(mag(i6*Lk2*(lambda2-lambda0).*(u1M-u0M)/xi^4+i6*Lk2*(lambda1-lambda0).*(u2M-u0M)/xi^4));

z220=intval(mag(i6*Lk2*(lambda1-lambda0).*(u1M-u0M)/xi^4));

z2=z200+z201+z202+z210+z211+z220;

z300=intval(mag(i3*Lk2.*u0M/xi^4))+i3*(int_cubicFFT(tuM,tuM,tuM)+i3*epsilon);

z301=intval(mag(i3*Lk2.*(u2M-u0M)/xi^4));

z310=intval(mag(i3*Lk2.*(u1M-u0M)/xi^4));

z3=z300+z301+z310;

tb2=[i1;i1;intval([1 (1:m-1)]').^(-intval(s))];

cubicZ1_M_m=abs(int_cubicFFT(abs(u0M),abs(u0M),[intval(zeros(m,1));tuM(m+1:end)]));

cubicZ1=int_cubicFFT(abs(u0M),abs(u0M),tuM);

Z_F=intval(zeros(M+2,3));

Z_F(1:m+2,1)=abs(intval(eye(m+2))-iAm*int_DF0M)*tb2 ... 
             +abs(iAm)*[abs(PHI_1-PHI_0)'*tb2+abs(PHI_2-PHI_0)'*tb2;i3*cubicZ1_M_m(1:m)+z1(1:m)+i3*epsilon(1:m)];
         
Z_F(m+3:end,1)=(cubicZ1(m+1:end)+i9*epsilon(m+1:end)+z1(m+1:end))./intval(mag(i1-(i1/lambda0)*Lk2(m+1:end)));

Z_F(:,2)=[abs(iAm)*[i0;i0;z2(1:m)];z2(m+1:end)./intval(mag(i1-(i1/lambda0)*Lk2(m+1:end)))];

Z_F(:,3)=[abs(iAm)*[i0;i0;z3(1:m)];z3(m+1:end)./intval(mag(i1-(i1/lambda0)*Lk2(m+1:end)))];

%%%%%%%%%%%%%%%%%
%%%% Z_infty %%%%
%%%%%%%%%%%%%%%%%

C1=iL^2*iM^2*lambda0/(iL^2*iM^2-lambda0);
mu_inv_M=lambda0/(iL^2*iM^2-lambda0);

z_inf_100=mu_inv_M*i3*A0^2*alphaM;

z_inf_101=C1*abs(lambda2-lambda0)/lambda0^2+i6*mu_inv_M*A0*A2_0*alphaM;

z_inf_102=abs(C1*abs(lambda2-lambda0)^2/xi^3)+i3*mu_inv_M*A2_0*A2_0*alphaM;

z_inf_110=C1*abs(lambda1-lambda0)/lambda0^2+i6*mu_inv_M*A0*A1_0*alphaM;

z_inf_111=abs(i2*C1*abs(lambda2-lambda0)*abs(lambda1-lambda0)/xi^3)+i6*mu_inv_M*A2_0*A1_0*alphaM;

z_inf_120=abs(C1*(lambda1-lambda0)^2/xi^3)+i3*mu_inv_M*A1_0*A1_0*alphaM;

z_inf_200=i6*mu_inv_M*A0*alphaM;

z_inf_201=i6*mu_inv_M*A2_0*alphaM;

z_inf_210=i6*mu_inv_M*A1_0*alphaM;

Z_infty(1,1)=(z_inf_100+z_inf_101+z_inf_102+z_inf_110+z_inf_111+z_inf_120)/iM^is;

Z_infty(1,2)=(z_inf_200+z_inf_201+z_inf_210)/iM^is;

Z_infty(1,3)=i3*mu_inv_M*alphaM/iM^is;

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Radii Polynomials %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%

omega=[1 1 1 (1:M).^(-s)];
p(1:M+2,:) = [sup(Y_F) sup(Z_F)];
p(M+3,:) = [0 sup(Z_infty)];
p(:,2)=p(:,2)-omega';  
In=[0,Inf];
k=-1;

while In(2)>In(1) && k<=M+1,
  R=roots(p(k+2,4:-1:1));
  if norm(imag(R))>0
      % S=[];
      % display(['Failure with the ' ,num2str(k+2) 'th radii polynomial.']);
      return
  else
      if max(R)<0
          % S=[];
          % display(['Failure with the ' ,num2str(k+2) 'th radii polynomial.']);
          return
      end
  [~,n_min]=min(R);
  R(n_min)=[];
  I_k(1)=min(R);
  I_k(2)=max(R);
  In(1)=max(In(1),I_k(1));
  In(2)=min(In(2),I_k(2));
  k=k+1;
  end
end

iomega=[i1 i1 i1 intval(1:M).^(-is)];
ip(1:M+2,:) = [Y_F Z_F];
ip(M+3,:) = [i0 Z_infty];
ip(:,2)=ip(:,2)-iomega'; 

if In(1)<In(2)       %%% Check that radii polynomials are really negative %%%
   r=min(1.1*In(1),(In(1)+In(2))/2);      
   pr=ip(:,4);
   for j=3:-1:1
     pr=ip(:,j)+intval(r)*pr;
   end
   if max(sup(pr))>=0
      return %, display('Failure! Empty intersection.')
   else
      % display('SUCCESS with the proof at the simplex!')
      valid=true;
   end
else
    return %, display('Failure! Empty interval.')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Verification of smoothness %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

int_s = infsup(0,1);
PSI_T = (PHI_0 + int_s*(PHI_1-PHI_0) + int_s*(PHI_2-PHI_0))';

M_smooth(:,1) = PSI_T*(bx1-bx0) - (PHI_1-PHI_0)'*r*infsup(-omega(1:m+2),omega(1:m+2))';
M_smooth(:,2) = PSI_T*(bx2-bx0) - (PHI_2-PHI_0)'*r*infsup(-omega(1:m+2),omega(1:m+2))';

det_M = M_smooth(1,1)*M_smooth(2,2) - M_smooth(1,2)*M_smooth(2,1);

if sign(inf(det_M)*sup(det_M))<0
  % display('Failure with the smoothness argument!')
  valid = false;
end

% S.decay=s;
% S.radius=r;
% S.In=In;

end

function [f]=int_f_ch(x)

i1=intval('1');
lambda=x(1);
c=x(2);
u=x(3:end);
m=length(u);
L=midrad(3.14159265358979323,1e-16);
f=(i1-(i1/lambda)*L^2*intval(0:m-1).^2)'.*u - int_cubicFFT(u,u,u);
f(1)=f(1)-c;

end

function [cubic]=int_cubicFFT(a1,a2,a3)
% the convolution term using FFT

i2=intval('2');
i4=intval('4');

M=size(a1,1);
% We make sure that the inputs are powers of 2 %
M1=(2^ceil(log(4*M-1)/log(2))-4*M)/2;   % Hence 4*M+2*M1 is a power of 2 %

ta1=[intval(zeros(M+M1+1,1));a1(M:-1:2,1);a1;intval(zeros(M+M1,1))];
ta2=[intval(zeros(M+M1+1,1));a2(M:-1:2,1);a2;intval(zeros(M+M1,1))];
ta3=[intval(zeros(M+M1+1,1));a3(M:-1:2,1);a3;intval(zeros(M+M1,1))];

tu1=intval_fft(ifftshift(ta1),intval(zeros(4*M+2*M1,1)),-1)/(i4*intval(M)+i2*intval(M1));
tu2=intval_fft(ifftshift(ta2),intval(zeros(4*M+2*M1,1)),-1)/(i4*intval(M)+i2*intval(M1));
tu3=intval_fft(ifftshift(ta3),intval(zeros(4*M+2*M1,1)),-1)/(i4*intval(M)+i2*intval(M1));

F=fftshift(intval_fft(tu1.*tu2.*tu3,intval(zeros(4*M+2*M1,1)),1));
cubic=(i4*intval(M)+i2*intval(M1))^i2*F(M1+2*M+1:M1+3*M,1);

end

function [X, Y] = intval_fft(x, y, sign)
% Computes the 1D inverse FFT of x+i*y. The length n of the
% vectors x and y MUST be a power of 2 (this is not checked for).

X = x(:); Y = y(:);

n = length(X);

% Implement bit-reversal part of algorithm
j = 0; n2 = n/2;
for i=0:n2-1
  if i < j                                      % First swap first half
    t1 = X(i+1); X(i+1) = X(j+1); X(j+1) = t1;  % Swap the real part
    t2 = Y(i+1); Y(i+1) = Y(j+1); Y(j+1) = t2;  % Swap the imaginary part

    if j < n2                                                     % Use mirrored effect on second half
      t1 = X(n-1-i+1); X(n-1-i+1) = X(n-1-j+1); X(n-1-j+1) = t1;  % Swap the real part
      t2 = Y(n-1-i+1); Y(n-1-i+1) = Y(n-1-j+1); Y(n-1-j+1) = t2;  % Swap the imaginary part
    end
  end

  m = n2;
  while ((m >= 2) && (m <= j))
    j = j - m;
    m = m/2;
  end

  j = j + m;
end

ipi=midrad(3.14159265358979323,1e-16);

% Now implement the Danielson-Lanczos algorithm
m = 1;

while (m < n) % Executed log2(n) times
 
  m2 = 2*m;
  theta = ipi/intval(m);     % Initialize trigonometric recurrence
  if sign==-1
      theta=-theta;
  end
  t1 = sin(intval('0.5')*theta);
  alpha = intval('-2')*t1*t1;
  beta = sin(theta);
  cs = intval('1'); sn = intval('0');
  
  for k=0:m-1
      
    t1=cs*X(k+m+1:m2:n+m)-sn*Y(k+m+1:m2:n+m);
    t2=cs*Y(k+m+1:m2:n+m)+sn*X(k+m+1:m2:n+m);
    
    X(k+m+1:m2:n+m)=X(k+1:m2:n)-t1;
    Y(k+m+1:m2:n+m)=Y(k+1:m2:n)-t2;
    
    X(k+1:m2:n)=X(k+1:m2:n)+t1;
    Y(k+1:m2:n)=Y(k+1:m2:n)+t2;

    % Trigonometric recurrence
    t1 = cs;
    cs = alpha*cs - beta*sn + cs;
    sn = alpha*sn + beta*t1 + sn;
  end
  
  m = 2*m;
end

X = [X(1); flipud(X(2:end))];
Y = [Y(1); flipud(Y(2:end))];

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Df] = int_Df_ch(x)

m=length(x)-2;
lambda=x(1);
u=x(3:end);
L=midrad(3.14159265358979323,1e-16);

Df_lambda=((L^2/lambda^2)*intval(0:m-1).^2)'.*u;
Df_c=intval([-1;zeros(m-1,1)]);
Df_u=int_f_ch_u(lambda,u);

Df = [Df_lambda Df_c Df_u];

end

function [J]=int_f_ch_u(lambda,a)
m=length(a);
L=midrad(3.14159265358979323,1e-16);
i1=intval('1');
i2=intval('2');
i3=intval('3');
i4=intval('4');

% We make sure that the inputs are powers of 2 %
m1=(2^ceil(log(4*m-1)/log(2))-4*m)/2;   % Hence 4*m+2*m1 is a power of 2 %

ta=[intval(zeros(m+m1+1,1));a(m:-1:2,1);a;intval(zeros(m+m1,1))];
tu=intval_fft(ifftshift(ta),intval(zeros(4*m+2*m1,1)),-1)/(i4*intval(m)+i2*intval(m1));

F=fftshift(intval_fft(tu.*tu,intval(zeros(4*m+2*m1,1)),1));
quad_sum=(i4*intval(m)+i2*intval(m1))*F(m+m1+2:4*m+m1-1,1)';
J=intval(zeros(m));

for j=0:m-1 
   J(j+1,1)=-i3*quad_sum(j+m);
   J(j+1,2:m)=-i3*(quad_sum(j+m-1:-1:j+1)+quad_sum(j+m+1:j+2*m-1));
   J(j+1,j+1)=J(j+1,j+1)+(i1-(i1/lambda)*L^2*j^2);
end

end